# frozen_string_literal: true

# checker Api::Dmo::HermesController avant d'ajouter des méthodes
class Hermes::Api::Dmo::UsersController < Hermes::Api::Dmo::ApplicationController
  hermes_tools 'aggregations'
  require Rails.root.join('lib', 'core_extentions', 'parameters', 'handle_sort').to_s

  VALID_OPERATORS = %w[
    $in
    $nin
    $ne
    $gt
    $gte
    $lt
    $lte
    $regex
  ].freeze

  ACCEPTABLE_PARAMS = %w[
   first_name
   last_name
   email
   gender
   phone_number
   auth_roles
  ].freeze

  REGEX_PARAMS = %w[
  ].freeze

  MUTATE_PARAMS = %i[
    first_name
    last_name
    email
    gender
    phone_number
    auth_roles
    entity_rights
    lessor_rights
  ].freeze
  mutate_params_only(*MUTATE_PARAMS)

  def initialize
    super
    @model = User
    @serializer = Hermes::UserSerializer
  end

  def check_for_beiam_error(user)
    user.instance_variable_get(:@create_user_response) && ![200, 201, 409].include?(user.instance_variable_get(:@create_user_response).code)
  end

  def beiam_error(user)
    {
      json: user.instance_variable_get(:@create_user_response).body,
      status: user.instance_variable_get(:@create_user_response).code
    }
  end

  def get_group_ids(p)
    error = nil
    group_ids = []

    [:entity_rights, :lessor_rights].each do |type|
      if p[type]
        p[type].each do |right|
          model = type == :entity_rights ? Hermes::Entity : Hermes::Lessor
          if a = model.find_by(id: right['id'])
            if @user.group_ids.include?(a.manager_group_id) || @user&.auth_roles&.include?('superadministrateur')
              group_id = nil
              case right['right']
              when 'administrator'
                group_id = a.manager_group_id
              when 'user'
                group_id = a.supervisor_group_id
              when 'reader'
                group_id = a.group_id
              end

              if group_id
                group_ids.push(group_id)
              end
            else
              error = {errors: "No #{type} on id : #{right['id']}"}
              break
            end
          else
            error = {errors: "Not found id : #{right['id']}"}
            break
          end
        end
      end
    end

    error || group_ids
  end

  def get_user_params(p, group_ids, user = nil)
    {
      username: p[:email] || user&.email,
      first_name: p[:first_name] || user&.first_name,
      last_name: p[:last_name] || user&.last_name,
      email: p[:email] || user&.email,
      gender: p[:gender] || user&.gender || 'M',
      phone_number: p[:phone_number] || user&.phone_number,
      login: p[:email] || user&.email,
      auth_roles: p[:auth_roles] || user&.auth_roles,
      group_ids: group_ids,
      provider: HERMES_CONFIG['identity_provider_key'],
      create_user_data: {
        gender: p[:gender] || user&.gender || 'M',
        lastName: p[:last_name] || user&.last_name,
        firstName: p[:first_name] || user&.first_name,
        email: p[:email] || user&.email,
        login: p[:email] || user&.email,
        mobileNumber: p[:phone_number] || user&.phone_number,
        roles: p[:auth_roles] || user&.auth_roles,
      },
    }
  end

  def can_create_user
    [
      'Administrator',
      'SuperAdmin',
      'Manager',
      'superadministrateur',
      'entity',
      'bailleur',
      'rdsl',
      'cre',
    ].any? { |role| current_user_roles&.include?(role) }
  end

  def can_create_user_with_auth_roles(p)
    return true if !p[:auth_roles]
    restricted_roles = [
      'Administrator',
      'SuperAdmin',
      'Manager',
      'superadministrateur',
    ]
    restricted_roles.all? { |role| !p[:auth_roles]&.include?(role) } || restricted_roles.any? { |role| current_user_roles&.include?(role) }
  end


  def create
    return render_json_error(:unauthorized, :cannot_create_user) unless can_create_user

    p = mutate_valid_params
    group_ids = get_group_ids(p)
    return render_json_error(:unauthorized, :no_right_on_entity) unless group_ids.kind_of?(Array)

    return render_json_error(:unauthorized, :cannot_create_user_with_auth_roles) unless can_create_user_with_auth_roles(p)

    user = User.create(get_user_params(p, group_ids).merge(password: Devise.friendly_token[0,20]))

    return render json: user.errors.messages, status: 400 if user.invalid?
    return render(beiam_error(user)) if check_for_beiam_error(user)

    render(jsonapi: user, serializer: @serializer, status: :created)
  end

  def set_instance
    @instance = @model.find_by(id: params[:id], **@query_precision) if params[:id]
  end

  def update
    return render_json_error(:not_found, :instance_not_found) unless @instance
    return render_json_error(:unauthorized, :cannot_create_user) unless can_create_user

    p = mutate_valid_params

    group_ids = get_group_ids(p)
    return render_json_error(:unauthorized, :no_right_on_entity) unless group_ids.kind_of?(Array)

    user_params = get_user_params(p, group_ids, @instance)
    @instance.update(user_params)

    return render json: @instance.errors.messages, status: 400 unless @instance.save
    return render(beiam_error(@instance)) if check_for_beiam_error(@instance)

    render(jsonapi: @instance, serializer: @serializer, status: :ok)
  end

  def index
    acceptable_params *ACCEPTABLE_PARAMS
    regex_params *REGEX_PARAMS
    valid_operators *VALID_OPERATORS
    super
  end

  def switch
    @target = User.find_by(id: params[:id])
    @target = User.find_by(main_applicant_id: params[:id]) || Hermes::Applicant.find_by(id: params[:id])&.user unless @target
    return render(json: { success: false, message: 'Utilisateur target non trouvé.' }, status: :not_found) unless @target
    return render nothing: true, status: :not_found unless @user

    if can_switch?
      if targeatable?
        success_switch_response
      else
        render json: {
          success: false,
          message: 'L\'utilisateur target doit être demandeur, bailleur ou entreprise.'
        }, status: :unprocessable_entity
      end
    else
      render json: {
        success: false,
        message: 'Vous n\'avez pas les droits nécessaire pour effectuer cette action.'
      }, status: :unauthorized
    end
  rescue ActionController::ParameterMissing => e
    render_json_error(:not_acceptable, :parameter_missing, { message: e.message })
  end

  private

  def can_switch?
    current_user_roles.include?('superadministrateur') || current_user_roles.include?('collaborateur offre logement')
  end

  def targeatable?
    return true if current_user_roles.include?('superadministrateur')

    (%w[demandeur bailleur entreprise] & @target.attributes.slice(:roles, :auth_roles).values.flatten.uniq.compact).any?
  end

  def success_switch_response
    token = @target.generate_jwt(true_user_id: @user.id.to_s)
    render json: {
      success: true,
      info: t(:registered),
      email: @target.email,
      login: @target.login,
      jwt_token: token
    }, status: :ok
  end
end
